@@ -5,16 +5,18 @@ use Test::Nginx::Socket -Base;
 my $code = $ENV{TEST_NGINX_INIT_BY_LUA};
 
 if ($code) {
-    $code =~ s/\\/\\\\/g;
-    $code =~ s/['"]/\\$&/g;
+    (my $escaped_code = $code) =~ s/\\/\\\\/g;
+    $escaped_code =~ s/['"]/\\$&/g;
 
     Test::Nginx::Socket::set_http_config_filter(sub {
         my $config = shift;
         if ($config =~ /init_by_lua_file/) {
             return $config;
         }
-        unless ($config =~ s{(?<!\#  )(?<!\# )(?<!\#)init_by_lua\s*(['"])((?:\\.|.)*)\1\s*;}{init_by_lua $1$code$2$1;}s) {
-            $config .= "init_by_lua '$code';";
+        unless ($config =~ s{(?<!\#  )(?<!\# )(?<!\#)init_by_lua\s*(['"])((?:\\.|.)*)\1\s*;}{init_by_lua $1$escaped_code$2$1;}s) {
+            unless ($config =~ s{(?<!\#  )(?<!\# )(?<!\#)init_by_lua_block\s*\{}{init_by_lua_block \{ $code }s) {
+                $config .= "init_by_lua '$escaped_code';";
+            }
         }
         return $config;
     });
@@ -70,7 +72,7 @@ Yichun "agentzh" Zhang (章亦春) C<< <agentzh@gmail.com> >>, CloudFlare Inc.
 
 =head1 COPYRIGHT & LICENSE
 
-Copyright (c) 2009-2014, Yichun Zhang C<< <agentzh@gmail.com> >>, CloudFlare Inc.
+Copyright (c) 2009-2016, Yichun Zhang C<< <agentzh@gmail.com> >>, CloudFlare Inc.
 
 This module is licensed under the terms of the BSD license.
 
