@@ -3,7 +3,7 @@ package Test::Nginx::Util;
 use strict;
 use warnings;
 
-our $VERSION = '0.24';
+our $VERSION = '0.25';
 
 use base 'Exporter';
 
@@ -20,6 +20,7 @@ use Scalar::Util qw( looks_like_number );
 use IO::Socket::INET;
 use IO::Socket::UNIX;
 use Test::LongString;
+use Carp qw( croak );
 
 our $ConfigVersion;
 our $FilterHttpConfig;
@@ -40,6 +41,8 @@ our $InSubprocess;
 our $RepeatEach = 1;
 our $MAX_PROCESSES = 10;
 
+our $LoadModules = $ENV{TEST_NGINX_LOAD_MODULES};
+
 our $NoShuffle = $ENV{TEST_NGINX_NO_SHUFFLE} || 0;
 
 our $UseValgrind = $ENV{TEST_NGINX_USE_VALGRIND};
@@ -232,7 +235,7 @@ our $DaemonEnabled          = 'on';
 our $ServerPort             = $ENV{TEST_NGINX_SERVER_PORT} || $ENV{TEST_NGINX_PORT} || 1984;
 our $ServerPortForClient    = $ENV{TEST_NGINX_CLIENT_PORT} || $ENV{TEST_NGINX_PORT} || 1984;
 our $NoRootLocation         = 0;
-our $TestNginxSleep         = $ENV{TEST_NGINX_SLEEP} || 0.05;
+our $TestNginxSleep         = $ENV{TEST_NGINX_SLEEP} || 0.015;
 our $BuildSlaveName         = $ENV{TEST_NGINX_BUILDSLAVE};
 our $ForceRestartOnTest     = (defined $ENV{TEST_NGINX_FORCE_RESTART_ON_TEST})
                                ? $ENV{TEST_NGINX_FORCE_RESTART_ON_TEST} : 1;
@@ -240,6 +243,19 @@ our $ForceRestartOnTest     = (defined $ENV{TEST_NGINX_FORCE_RESTART_ON_TEST})
 our $ChildPid;
 our $UdpServerPid;
 our $TcpServerPid;
+our @EnvToNginx;
+
+sub env_to_nginx (@) {
+    if (!@_) {
+        croak "env_to_nginx: no arguments specified";
+    }
+    for my $v (@_) {
+        if ($v !~ /^[A-Za-z_]/ || $v =~ /\n/) {
+            croak "env_to_nginx: bad argument value: $v\n";
+        }
+        push @EnvToNginx, $v;
+    }
+}
 
 sub sleep_time {
     return $TestNginxSleep;
@@ -271,9 +287,9 @@ sub repeat_each (@) {
             return;
         }
         $RepeatEach = shift;
-    } else {
-        return $RepeatEach;
     }
+
+    return $RepeatEach;
 }
 
 sub worker_connections (@) {
@@ -333,6 +349,7 @@ sub master_process_enabled (@) {
 }
 
 our @EXPORT = qw(
+    env_to_nginx
     is_str
     check_accum_error_log
     is_running
@@ -410,7 +427,6 @@ our $CheckErrorLog;
 
 our $NginxVersion;
 our $NginxRawVersion;
-our $TODO;
 
 sub add_block_preprocessor(&) {
     unshift @BlockPreprocessors, shift;
@@ -506,8 +522,8 @@ sub kill_process ($$) {
             warn "WARNING: killing the child process $pid.\n";
         }
 
-        if (kill(SIGQUIT, $pid) == 0) { # send quit signal
-            warn "WARNING: failed to send quit signal to the child process with PID $pid.\n";
+        if (kill(SIGTERM, $pid) == 0) { # send term signal
+            warn "WARNING: failed to send term signal to the child process with PID $pid.\n";
         }
 
         sleep $TestNginxSleep * $i;
@@ -516,6 +532,7 @@ sub kill_process ($$) {
         $i++;
     }
 
+    #system("ps aux|grep $pid > /dev/stderr");
     warn "WARNING: killing the child process $pid with force...";
 
     kill(SIGKILL, $pid);
@@ -579,7 +596,7 @@ sub run_tests () {
 
     for my $block ($NoShuffle ? Test::Base::blocks() : shuffle Test::Base::blocks()) {
         for my $hdl (@BlockPreprocessors) {
-            $block = $hdl->($block);
+            $hdl->($block);
         }
         run_test($block);
     }
@@ -766,21 +783,34 @@ sub write_config_file ($$) {
         undef $PostponeOutput;
     }
 
+    my $extra_http_config = '';
+
     if (defined $PostponeOutput) {
         if ($PostponeOutput !~ /^\d+$/) {
             bail_out "Bad TEST_NGINX_POSTPOHNE_OUTPUT value: $PostponeOutput\n";
         }
-        $http_config .= "\n    postpone_output $PostponeOutput;\n";
+        $extra_http_config .= "\n    postpone_output $PostponeOutput;\n";
     }
 
     if (!defined $main_config) {
         $main_config = '';
     }
 
+    if ($LoadModules) {
+        my @modules = map { "load_module $_;" } grep { $_ } split /\s+/, $LoadModules;
+        if (@modules) {
+            $main_config .= join " ", @modules;
+        }
+    }
+
+    $main_config = expand_env_in_config($main_config);
+
     if (!defined $post_main_config) {
         $post_main_config = '';
     }
 
+    $post_main_config = expand_env_in_config($post_main_config);
+
     if ($CheckLeak || $Benchmark) {
         $LogLevel = 'warn';
         $AccLogFile = 'off';
@@ -811,6 +841,17 @@ env MOCKEAGAIN_WRITE_TIMEOUT_PATTERN;
 env LD_PRELOAD;
 env LD_LIBRARY_PATH;
 env DYLD_INSERT_LIBRARIES;
+env DYLD_FORCE_FLAT_NAMESPACE;
+_EOC_
+
+    for my $v (@EnvToNginx) {
+        if ($v =~ /\s/) {
+            $v = "'$v'";
+        }
+        print $out "env $v;\n";
+    }
+
+    print $out <<_EOC_;
 #env LUA_PATH;
 #env LUA_CPATH;
 
@@ -863,6 +904,8 @@ _EOC_
             return 200 '$ConfigVersion';
         }
     }
+
+$extra_http_config
 _EOC_
     }
 
@@ -887,6 +930,10 @@ _EOC_
 
     print $out "}\n";
 
+    print $out <<_EOC_;
+env ASAN_OPTIONS;
+_EOC_
+
     close $out;
 }
 
@@ -897,14 +944,13 @@ sub get_canon_version (@) {
 sub get_nginx_version () {
     my $out = `$NginxBinary -V 2>&1`;
     if (!defined $out || $? != 0) {
-        warn "Failed to get the version of the Nginx in PATH.\n";
+        bail_out("Failed to get the version of the Nginx in PATH");
     }
     if ($out =~ m{(?:nginx|openresty)/(\d+)\.(\d+)\.(\d+)}s) {
         $NginxRawVersion = "$1.$2.$3";
         return get_canon_version($1, $2, $3);
     }
-    warn "Failed to parse the output of \"nginx -V\": $out\n";
-    return undef;
+    bail_out("Failed to parse the output of \"nginx -V\": $out\n");
 }
 
 sub get_pid_from_pidfile ($) {
@@ -1321,6 +1367,22 @@ sub run_test ($) {
         }
     }
 
+    my $todo = $block->todo;
+    if (defined $todo) {
+        if ($todo =~ m{
+                ^ \s* (\d+) \s* : \s* (.*)
+            }xs)
+        {
+            $should_todo = 1;
+            $tests_to_skip = $1;
+            $todo_reason = $2;
+        } else {
+            bail_out("$name - Invalid --- todo spec: " .
+                $todo);
+            die;
+        }
+    }
+
     if (!defined $todo_reason) {
         $todo_reason = "various reasons";
     }
@@ -1637,18 +1699,20 @@ start_nginx:
 RUN_AGAIN:
                 system($cmd);
 
-                if ($? == -1) {
+                my $status = $?;
+
+                if ($status == -1) {
                     $exec_failed = 1;
 
                 } else {
-                    $exit_code = $? >> 8;
+                    $exit_code = $status >> 8;
 
                     if ($? > (128 << 8)) {
                         $coredump = ($exit_code & 128);
                         $exit_code = ($exit_code >> 8);
 
                     } else {
-                        $coredump = ($? & 128);
+                        $coredump = ($status & 128);
                     }
                 }
 
@@ -1665,23 +1729,38 @@ RUN_AGAIN:
                                        "$name - die with the expected exit code")
 
                     } else {
-                        Test::More::isnt($?, 0, "$name - die as expected")
+                        Test::More::isnt($status, 0, "$name - die as expected")
                     }
 
                     $CheckErrorLog->($block, undef, $dry_run, $i, 0);
 
+                    #warn "Status: $status\n";
+                    if ($status == 0) {
+                        warn("WARNING: $name - nginx must die but it does ",
+                             "not; killing it (req $i)");
+                        my $max_i = 15;
+                        for (my $i = 1; $i <= $max_i; $i++) {
+                            if (-f $PidFile) {
+                                last;
+                            }
+                            sleep $TestNginxSleep;
+                        }
+                        my $pid = get_pid_from_pidfile($name);
+                        kill_process($pid, 1);
+                    }
+
                     goto RUN_AGAIN if ++$i < $RepeatEach;
                     return;
                 }
 
-                if ($? != 0) {
+                if ($status != 0) {
                     if ($ENV{TEST_NGINX_IGNORE_MISSING_DIRECTIVES} and
                             my $directive = check_if_missing_directives())
                     {
                         $dry_run = "the lack of directive $directive";
 
                     } else {
-                        bail_out("$name - Cannot start nginx using command \"$cmd\".");
+                        bail_out("$name - Cannot start nginx using command \"$cmd\" (status code $status).");
                     }
                 }
             }
@@ -2042,7 +2121,11 @@ request:
                 }
 
                 my $buf = '';
-                $udp_socket->recv($buf, 4096);
+                my $sender = $udp_socket->recv($buf, 4096);
+                #warn "sender: $sender";
+                if (!defined $sender) {
+                    warn "udp recv failed: $!";
+                }
 
                 if ($Verbose) {
                     warn "UDP server has got data: ", length $buf, "\n";
@@ -2063,7 +2146,7 @@ request:
                 my $delay = parse_time($block->udp_reply_delay);
                 if ($delay) {
                     if ($Verbose) {
-                        warn "sleep $delay before sending UDP reply\n";
+                        warn "$name - sleep $delay before sending UDP reply\n";
                     }
                     sleep $delay;
                 }
@@ -2077,21 +2160,25 @@ request:
 
                     if ($ref) {
                         if ($ref ne 'ARRAY') {
-                            bail_out("bad --- udp_reply value");
+                            warn("$name - Bad --- udp_reply value");
                         }
 
                         for my $r (@$reply) {
                             #warn "sending reply $r";
                             my $bytes = $udp_socket->send($r);
                             if (!defined $bytes) {
-                                warn "WARNING: udp server failed to send reply: $!\n";
+                                warn "$name - WARNING: udp server failed to send reply: $!\n";
                             }
                         }
 
                     } else {
+                        if ($reply =~ /syntax error at \(eval \d+\) line \d+, near/) {
+                            bail_out("$name - Bad --- udp_reply: $reply");
+                        }
+
                         my $bytes = $udp_socket->send($reply);
                         if (!defined $bytes) {
-                            warn "WARNING: udp server failed to send reply: $!\n";
+                            warn "$name - WARNING: udp server failed to send reply: $!\n";
                         }
                     }
                 }
@@ -2127,7 +2214,7 @@ request:
 
         } elsif ($should_todo) {
             TODO: {
-                local $TODO = "$name - $todo_reason";
+                Test::More::todo_skip("$name - $todo_reason", $tests_to_skip);
 
                 $RunTestHelper->($block, $dry_run, $i - 1);
                 run_tcp_server_tests($block, $tcp_socket, $tcp_query_file);
@@ -2226,16 +2313,18 @@ retry:
                     kill(SIGKILL, $pid);
                     waitpid($pid, 0);
 
-                    unlink $PidFile or
-                        bail_out "Failed to remove pid file $PidFile\n";
+                    if (!unlink($PidFile) && -f $PidFile) {
+                        bail_out "Failed to remove pid file $PidFile: $!\n";
+                    }
 
                 } else {
                     #warn "nginx killed";
                 }
 
             } else {
-                unlink $PidFile or
-                    bail_out "Failed to remove pid file $PidFile\n";
+                if (!unlink($PidFile) && -f $PidFile) {
+                    bail_out "Failed to remove pid file $PidFile: $!\n";
+                }
             }
         } else {
             #warn "pid file not found";
